from abnf.parser import Rule as _Rule

from . import rfc3986
from .misc import load_grammar


@load_grammar(
    [
        ("scheme", rfc3986.Rule("scheme")),
        ("port", rfc3986.Rule("port")),
        ("IP-literal", rfc3986.Rule("IP-literal")),
        ("IPvFuture", rfc3986.Rule("IPvFuture")),
        ("IPv6address", rfc3986.Rule("IPv6address")),
        ("h16", rfc3986.Rule("h16")),
        ("ls32", rfc3986.Rule("ls32")),
        ("IPv4address", rfc3986.Rule("IPv4address")),
        ("dec-octet", rfc3986.Rule("dec-octet")),
        ("pct-encoded", rfc3986.Rule("pct-encoded")),
        ("unreserved", rfc3986.Rule("unreserved")),
        ("reserved", rfc3986.Rule("reserved")),
        ("gen-delims", rfc3986.Rule("gen-delims")),
        ("sub-delims", rfc3986.Rule("sub-delims")),
    ]
)
class Rule(_Rule):
    """Grammar from RFC 3987."""

    grammar = """
IRI            = scheme ":" ihier-part [ "?" iquery ]
                    [ "#" ifragment ]

ihier-part     = "//" iauthority ipath-abempty
                / ipath-absolute
                / ipath-rootless
                / ipath-empty

IRI-reference  = IRI / irelative-ref

absolute-IRI   = scheme ":" ihier-part [ "?" iquery ]

irelative-ref  = irelative-part [ "?" iquery ] [ "#" ifragment ]

irelative-part = "//" iauthority ipath-abempty
                / ipath-absolute
                / ipath-noscheme
                / ipath-empty

iauthority     = [ iuserinfo "@" ] ihost [ ":" port ]
iuserinfo      = *( iunreserved / pct-encoded / sub-delims / ":" )
ihost          = IP-literal / IPv4address / ireg-name

ireg-name      = *( iunreserved / pct-encoded / sub-delims )

ipath          = ipath-abempty   ; begins with "/" or is empty
                / ipath-absolute  ; begins with "/" but not "//"
                / ipath-noscheme  ; begins with a non-colon segment
                / ipath-rootless  ; begins with a segment
                / ipath-empty     ; zero characters

ipath-abempty  = *( "/" isegment )
ipath-absolute = "/" [ isegment-nz *( "/" isegment ) ]
ipath-noscheme = isegment-nz-nc *( "/" isegment )
ipath-rootless = isegment-nz *( "/" isegment )
ipath-empty    = 0<ipchar>

isegment       = *ipchar
isegment-nz    = 1*ipchar
isegment-nz-nc = 1*( iunreserved / pct-encoded / sub-delims
                    / "@" )
                ; non-zero-length segment without any colon ":"

ipchar         = iunreserved / pct-encoded / sub-delims / ":"
                / "@"

iquery         = *( ipchar / iprivate / "/" / "?" )

ifragment      = *( ipchar / "/" / "?" )

iunreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~" / ucschar

ucschar        = %xA0-D7FF / %xF900-FDCF / %xFDF0-FFEF
                / %x10000-1FFFD / %x20000-2FFFD / %x30000-3FFFD
                / %x40000-4FFFD / %x50000-5FFFD / %x60000-6FFFD
                / %x70000-7FFFD / %x80000-8FFFD / %x90000-9FFFD
                / %xA0000-AFFFD / %xB0000-BFFFD / %xC0000-CFFFD
                / %xD0000-DFFFD / %xE1000-EFFFD

iprivate       = %xE000-F8FF / %xF0000-FFFFD / %x100000-10FFFD
    """


# ßee https://www.rfc-editor.org/rfc/rfc3987#section-2.2
for rule in Rule.rules():
    rule.first_match_alternation = True



import typing

import pytest

from abnf.grammars.misc import load_grammar, load_grammar_rules
from abnf.parser import Literal
from abnf.parser import Rule as _Rule


class ImportRule(_Rule):
    pass


ImportRule("test", Literal("test"))


@load_grammar_rules([("test", ImportRule("test"))])
class Rule(_Rule):
    grammar: typing.List[str] = []


def test_misc_load_grammar_rules_import():
    assert Rule("test").definition == ImportRule("test").definition


@load_grammar([("test", ImportRule("test"))])
class Rule1(_Rule):
    grammar: str = ''

def test_load_grammar():
    assert Rule("test").definition == ImportRule("test").definition


class Foo(_Rule):
    grammar = 'foo="bar"'


def test_load_grammar_rules_str():
    with pytest.raises(TypeError):
        load_grammar_rules()(Foo)


class Bar(_Rule):
    grammar = ['foo="bar"']


def test_load_grammar_list():
    with pytest.raises(TypeError):
        load_grammar()(Bar)



import pytest

from abnf.grammars import rfc6266


@pytest.mark.parametrize(
    "src",
    [
        "Content-Disposition: inline",
        "Content-Disposition: attachment",
        'Content-Disposition: attachment; filename="filename.jpg"',
        'Content-Disposition: form-data; name="fieldName"',
        'Content-Disposition: form-data; name="fieldName"; filename="filename.jpg"',
        "Content-Disposition: attachment; filename*=UTF-8''%e2%82%ac%20rates",
        "Content-Disposition: attachment; filename=\"EURO rates\"; filename*=utf-8''%e2%82%ac%20rates",
    ],
)
def test_content_disposition(src: str):
    content_disposition = rfc6266.Rule("content-disposition")
    assert content_disposition.parse_all(src)


def test_disp_ext_parm():
    # this test checks to ensure that the RFC 6266 grammar as modified in response to erratum
    # parses src into an ext-name = ext-value parameter.
    src = "foo*=utf-8''%E2%88%Aar"
    node = rfc6266.Rule("disp-ext-parm").parse_all(src)
    assert len(node.children) == 3
    assert node.children[0].name == "ext-parmname"
    assert node.children[0].value == "foo*"
    assert node.children[1].value == "="
    assert node.children[2].name == "ext-value"
    assert node.children[2].value == "utf-8''%E2%88%Aar"



from abnf.grammars import rfc3987


def test_ipath_empty():
    # ipath-empty definition uses the prose-val as rulename thing,
    # so we check to ensure that parsing succeeds.

    src = ""
    node = rfc3987.Rule("ipath-empty").parse_all(src)
    assert node.value == src



"""
Collected rules from RFC 2616.
https://www.rfc-editor.org/rfc/rfc2616.html

Though RFC 2616 is largely obsolete, some later RFCs, not obsolete, 
still incorporate grammar from RFC 2616; in particular, RFC 6265.
Here we collect grammar from RFC 2616 as needed for use with other RFC grammars.
"""


from abnf.parser import Rule as _Rule
from .misc import load_grammar_rules


@load_grammar_rules()
class Rule(_Rule):
    """Rule objects generated from ABNF in RFC 5234."""

    grammar = []



"""
Collected rules from RFC 9110
https://www.rfc-editor.org/rfc/rfc9110.html
"""

from abnf.parser import Rule as _Rule

from . import rfc3986, rfc4647, rfc5322, rfc5646
from .misc import load_grammar


@load_grammar(
    [
        ("language-range", rfc4647.Rule("language-range")),
        ("mailbox", rfc5322.Rule("mailbox")),
        ("language-tag", rfc5646.Rule("language-tag")),
        ("URI-reference", rfc3986.Rule("URI-reference")),
        ("absolute-URI", rfc3986.Rule("absolute-URI")),
        ("relative-part", rfc3986.Rule("relative-part")),
        ("segment", rfc3986.Rule("segment")),
        ("authority", rfc3986.Rule("authority")),
        ("port", rfc3986.Rule("port")),
        ("query", rfc3986.Rule("query")),
        ("uri-host", rfc3986.Rule("host")),
        ("path-abempty", rfc3986.Rule("path-abempty")),
    ]
)
class Rule(_Rule):
    """Rules from RFC 9110."""

    grammar = r"""
Accept = [ ( media-range [ weight ] ) *( OWS "," OWS ( media-range [
 weight ] ) ) ]
Accept-Charset = [ ( ( token / "*" ) [ weight ] ) *( OWS "," OWS ( (
 token / "*" ) [ weight ] ) ) ]
Accept-Encoding = [ ( codings [ weight ] ) *( OWS "," OWS ( codings [
 weight ] ) ) ]
Accept-Language = [ ( language-range [ weight ] ) *( OWS "," OWS (
 language-range [ weight ] ) ) ]
Accept-Ranges = acceptable-ranges
Allow = [ method *( OWS "," OWS method ) ]
Authentication-Info = [ auth-param *( OWS "," OWS auth-param ) ]
Authorization = credentials

BWS = OWS

Connection = [ connection-option *( OWS "," OWS connection-option )
 ]
Content-Encoding = [ content-coding *( OWS "," OWS content-coding )
 ]
Content-Language = [ language-tag *( OWS "," OWS language-tag ) ]
Content-Length = 1*DIGIT
Content-Location = absolute-URI / partial-URI
Content-Range = range-unit SP ( range-resp / unsatisfied-range )
Content-Type = media-type

Date = HTTP-date

ETag = entity-tag
Expect = [ expectation *( OWS "," OWS expectation ) ]

From = mailbox

GMT = %x47.4D.54 ; GMT

HTTP-date = IMF-fixdate / obs-date
Host = uri-host [ ":" port ]

IMF-fixdate = day-name "," SP date1 SP time-of-day SP GMT
If-Match = "*" / [ entity-tag *( OWS "," OWS entity-tag ) ]
If-Modified-Since = HTTP-date
If-None-Match = "*" / [ entity-tag *( OWS "," OWS entity-tag ) ]
If-Range = entity-tag / HTTP-date
If-Unmodified-Since = HTTP-date

Last-Modified = HTTP-date
Location = URI-reference

Max-Forwards = 1*DIGIT

OWS = *( SP / HTAB )

Proxy-Authenticate = [ challenge *( OWS "," OWS challenge ) ]
Proxy-Authentication-Info = [ auth-param *( OWS "," OWS auth-param )
 ]
Proxy-Authorization = credentials

RWS = 1*( SP / HTAB )
Range = ranges-specifier
Referer = absolute-URI / partial-URI
Retry-After = HTTP-date / delay-seconds

Server = product *( RWS ( product / comment ) )

TE = [ t-codings *( OWS "," OWS t-codings ) ]
Trailer = [ field-name *( OWS "," OWS field-name ) ]

URI-reference = <URI-reference, see [URI], Section 4.1>
Upgrade = [ protocol *( OWS "," OWS protocol ) ]
User-Agent = product *( RWS ( product / comment ) )

Vary = [ ( "*" / field-name ) *( OWS "," OWS ( "*" / field-name ) )
 ]
Via = [ ( received-protocol RWS received-by [ RWS comment ] ) *( OWS
 "," OWS ( received-protocol RWS received-by [ RWS comment ] ) ) ]

WWW-Authenticate = [ challenge *( OWS "," OWS challenge ) ]

absolute-URI = <absolute-URI, see [URI], Section 4.3>
absolute-path = 1*( "/" segment )
acceptable-ranges = range-unit *( OWS "," OWS range-unit )
asctime-date = day-name SP date3 SP time-of-day SP year
auth-param = token BWS "=" BWS ( token / quoted-string )
auth-scheme = token
authority = <authority, see [URI], Section 3.2>

challenge = auth-scheme [ 1*SP ( token68 / [ auth-param *( OWS ","
 OWS auth-param ) ] ) ]
codings = content-coding / "identity" / "*"
comment = "(" *( ctext / quoted-pair / comment ) ")"
complete-length = 1*DIGIT
connection-option = token
content-coding = token
credentials = auth-scheme [ 1*SP ( token68 / [ auth-param *( OWS ","
 OWS auth-param ) ] ) ]
ctext = HTAB / SP / %x21-27 ; \'!\'-\''\'
 / %x2A-5B ; \'*\'-\'[\'
 / %x5D-7E ; \']\'-\'~\'
 / obs-text

date1 = day SP month SP year
date2 = day "-" month "-" 2DIGIT
date3 = month SP ( 2DIGIT / ( SP DIGIT ) )
day = 2DIGIT
day-name = %x4D.6F.6E ; Mon
 / %x54.75.65 ; Tue
 / %x57.65.64 ; Wed
 / %x54.68.75 ; Thu
 / %x46.72.69 ; Fri
 / %x53.61.74 ; Sat
 / %x53.75.6E ; Sun
day-name-l = %x4D.6F.6E.64.61.79 ; Monday
 / %x54.75.65.73.64.61.79 ; Tuesday
 / %x57.65.64.6E.65.73.64.61.79 ; Wednesday
 / %x54.68.75.72.73.64.61.79 ; Thursday
 / %x46.72.69.64.61.79 ; Friday
 / %x53.61.74.75.72.64.61.79 ; Saturday
 / %x53.75.6E.64.61.79 ; Sunday
delay-seconds = 1*DIGIT

entity-tag = [ weak ] opaque-tag
etagc = "!" / %x23-7E ; \'#\'-\'~\'
 / obs-text
expectation = token [ "=" ( token / quoted-string ) parameters ]

field-content = field-vchar [ 1*( SP / HTAB / field-vchar )
 field-vchar ]
field-name = token
field-value = *field-content
field-vchar = VCHAR / obs-text
first-pos = 1*DIGIT

hour = 2DIGIT
http-URI = "http://" authority path-abempty [ "?" query ]
https-URI = "https://" authority path-abempty [ "?" query ]

incl-range = first-pos "-" last-pos
int-range = first-pos "-" [ last-pos ]

language-range = <language-range, see [RFC4647], Section 2.1>
language-tag = <Language-Tag, see [RFC5646], Section 2.1>
last-pos = 1*DIGIT

mailbox = <mailbox, see [RFC5322], Section 3.4>
media-range = ( "*/*" / ( type "/*" ) / ( type "/" subtype ) )
 parameters
media-type = type "/" subtype parameters
method = token
minute = 2DIGIT
month = %x4A.61.6E ; Jan
 / %x46.65.62 ; Feb
 / %x4D.61.72 ; Mar
 / %x41.70.72 ; Apr
 / %x4D.61.79 ; May
 / %x4A.75.6E ; Jun
 / %x4A.75.6C ; Jul
 / %x41.75.67 ; Aug
 / %x53.65.70 ; Sep
 / %x4F.63.74 ; Oct
 / %x4E.6F.76 ; Nov
 / %x44.65.63 ; Dec

obs-date = rfc850-date / asctime-date
obs-text = %x80-FF
opaque-tag = DQUOTE *etagc DQUOTE
other-range = 1*( %x21-2B ; \'!\'-\'+\'
 / %x2D-7E ; \'-\'-\'~\'
 )

parameter = parameter-name "=" parameter-value
parameter-name = token
parameter-value = ( token / quoted-string )
parameters = *( OWS ";" OWS [ parameter ] )
partial-URI = relative-part [ "?" query ]
path-abempty = <path-abempty, see [URI], Section 3.3>
port = <port, see [URI], Section 3.2.3>
product = token [ "/" product-version ]
product-version = token
protocol = protocol-name [ "/" protocol-version ]
protocol-name = token
protocol-version = token
pseudonym = token

qdtext = HTAB / SP / "!" / %x23-5B ; \'#\'-\'[\'
 / %x5D-7E ; \']\'-\'~\'
 / obs-text
query = <query, see [URI], Section 3.4>
quoted-pair = "\" ( HTAB / SP / VCHAR / obs-text )
quoted-string = DQUOTE *( qdtext / quoted-pair ) DQUOTE
qvalue = ( "0" [ "." *3DIGIT ] ) / ( "1" [ "." *3"0" ] )

range-resp = incl-range "/" ( complete-length / "*" )
range-set = range-spec *( OWS "," OWS range-spec )
range-spec = int-range / suffix-range / other-range
range-unit = token
ranges-specifier = range-unit "=" range-set
received-by = pseudonym [ ":" port ]
received-protocol = [ protocol-name "/" ] protocol-version
relative-part = <relative-part, see [URI], Section 4.2>
rfc850-date = day-name-l "," SP date2 SP time-of-day SP GMT

second = 2DIGIT
segment = <segment, see [URI], Section 3.3>
subtype = token
suffix-length = 1*DIGIT
suffix-range = "-" suffix-length

t-codings = "trailers" / ( transfer-coding [ weight ] )
tchar = "!" / "#" / "$" / "%" / "&" / "'" / "*" / "+" / "-" / "." /
 "^" / "_" / "`" / "|" / "~" / DIGIT / ALPHA
time-of-day = hour ":" minute ":" second
token = 1*tchar
token68 = 1*( ALPHA / DIGIT / "-" / "." / "_" / "~" / "+" / "/" )
 *"="
transfer-coding = token *( OWS ";" OWS transfer-parameter )
transfer-parameter = token BWS "=" BWS ( token / quoted-string )
type = token

unsatisfied-range = "*/" complete-length
uri-host = <host, see [URI], Section 3.2.2>

weak = %x57.2F ; W/
weight = OWS ";" OWS "q=" qvalue

year = 4DIGIT
"""



import pytest

from abnf.grammars import rfc6265


@pytest.mark.parametrize("src", [
    'bar',
    '"bar"',
    ])
def test_cookie_value(src: str):
    rfc6265.Rule('cookie-value').parse_all(src)


@pytest.mark.parametrize("src", [
    "www.example.com",
    "127.0.0.1",
    "FE80:CD00:0000:0CDE:1257:0000:211E:729C"
])
def test_domain_value(src: str):
    node = rfc6265.Rule('domain-value').parse_all(src)
    assert node.value == src